{ $DEFINE DEBUG}
// -----------------------------------------------------------------------------
// Copyright  1998 - 2000 Andy Radchenko. All Rights (and Lefts) Reserved
// -----------------------------------------------------------------------------
// This is my second attempt of the implementation of the HP-GL/2 interpreter
// First attempt handle simple hpgl files pretty well, however, a more complex
// files did not come up properly. Everybody is invited to improve this component
// It will work with Delphi 5 ONLY! HP_Elem unit, responsible for collection and
// drawing of the plot file elements, is using the TObjectList, introduced in the
// Delphi 5. If you do not use latest version of Delphi, you have to invent this
// list or use 3rd party component, similar in functionality to the TObjectList.
// -----------------------------------------------------------------------------
// Please send your comments to arussian@prodigy.net or andyr@reprografia.com
//------------------------------------------------------------------------------
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the author be held liable for any  damages arising from the
// use of this software.

// This software is distributed as a FreeWare, however author reservs rights
// to change it at any time. You can use it for any purpose, including commercial
// applications, and to alter it and redistribute it freely, subject to the
// following restrictions:
//
// A. The origin of this software must not be misrepresented and you shall not
//    claim that you have written the original code. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    highly appreciated, however it is not required.
//
// B. Redacted source code must be plainly marked as such, and must not be
//    misrepresented as being a part of the original code.
//
// C. This notice may not be removed or altered from any source distribution.
//------------------------------------------------------------------------------
{$A+}
unit hpgl_2;
interface
uses
  Classes, Windows, Graphics, Controls, Messages, Dialogs, SysUtils,
  Printers, HP_Elem, ExtCtrls;

const
  hpMaxParams = 127;
  hpMaxPens   = 255;
  Delta       = 5 * Pi / 180;  // Arc angle increment in radians
  C_to_R      = 2 * Pi;        // Circle sweep in radians
  Pi_090      = Pi / 2;        // 90 degrees in radians
  Pi_270      = Pi * 1.5;      // 270 degrees in radians
  Pi_360      = Pi * 2;        // 360 degrees in radians
// Error codes reported by HPGL/2
  HpErrorBase   = 1200; // Use this number to get a string from a string table
  errHpOk       = 00;
  errHpNList    = 01;   // Error creating node list
  errHpEList    = 02;   // Error creating element list
  errHpMStrm    = 03;   // Error creating memory stream
  errHpFName    = 04;   // File name is not assigned
  errNoFile     = 05;   // File not found
  errHpAddPnt   = 06;   // Error adding point to a temp storage
  errHpAddElm   = 07;   // Error adding HP element
  errHpAddImg   = 08;   // Error adding image
  errHpInvPln   = 09;   // Invalid data in the Polyline Encode
  errHpInvHpCmd = 10;   // Invalid or unrecognised HP command
  errHpInvRtCmd = 11;   // Invalid RTL command
  errHpInvPJCmd = 12;   // Invalid PJL command
  errHpNoCanvas = 13;   // Canvas property is not assigned;
  errHpNoParent = 14;   // Parent property is not assigned;
  errHpNoImgMem = 15;   // Not enough memory for the bitmaps
  errHpError    = 16;
  errRTLError   = 17;
  errPJLError   = 18;
  errZeroSize   = 19;   // File length is 0 bytes
  errIncomplete = 20;
  {}
type
  THpBmpData = record             // HP RTL bitmap parameters
                 ClrSpace,        // Color space, allways 0
                 Encode,          // Pixel encoding
                 BitsNdx,         // Number of bits per index
                 BitsRed,         // Number of bits per primary red
                 BitsGrn,         // Number of bits per primary green
                 BitsBlu  : byte; // Number of bits per primary blue
                 WRed,            // White reference for red
                 WGrn,            // White reference for green
                 WBlu,            // White reference for blue
                 BRed,            // Black reference for red
                 BGrn,            // Black reference for green
                 BBlu     : word; // Black reference for blue
               end;
  {}
  THpPenData  = record
                  Color : TColor;
                  Width : real;
                end;
  THpPenArray = array[1..hpMaxPens] of THpPenData;
  {}
  TPointPtr  = ^TPoint;
  TRealPoint = record
                 X, Y : real;
               end;
  TCharPtr  = ^Char;
  {}
  THPGLError = procedure(Sender : TObject; Code : integer;
                         MoreInfo : string) of object;
  THPGLErrPtr = ^THPGLError;
  {}
  THPGLInfo  = procedure(Sender : TObject; Count : integer) of object;
  THPGLInfoPtr  = ^THPGLInfo;
  {}
  THPGLProgr = procedure(Sender : TObject; Percent : integer) of object;
  THPGLProgPtr = ^THPGLProgr;
  {}
  TUseArray = array[0..hpMaxPens] of Boolean;
  {}
  TViewMode = (vmZoom, vmPan);
  {}
  THPGL2 = class(TGraphicControl)
  private
    fOnPaint,
    fOnReady,
    fAfterPaint : TNotifyEvent;
    {}
    fPlotFile   : string;         // Plot file name
    fStream     : TMemoryStream;
    fPage,                        // Corners of the plot page
    fHPPage,                      // Corners of the physical page in HPGL
    fPJPage,                      // Corners of the physical page in PJL
    fRTPage,                      // Corners of the physical page in RTL
    fLastPos,
    fPenPos,                      // Current pen position in HPGL
    fRtlCap     : TPoint;         // -- " -- " -- " -- " -- RTL
    fRealP1,                      // Corner of the page in the device units
    fRealP2,
    fUserP1,                      // Corners of the page in user's coords
    fUserP2     : TRealPoint;
    fRTLScale,
    fScaleX,                      // UCS scale
    fScaleY     : real;
    fWinList,
    fNodeList,                    // polyline & polygon buffers
    fPolyList   : THpNodeList;
    fFrame,
    fPaper      : TColor;          // Background color and outline;
    fRTLSPal,                      // Push buffer for RTL Palette
    fRTLPal     : array[0..255] of TColor; // RTL palette
    fPens       : THpPenArray; // Global pen table
    fPenUsed    : TUseArray;
    fOldBytes,
    fOldCount,
    fPenCount,
    fRTLWidth,                    // Width & Heigth of the current RTL bitmap
    fRTLHeight,
    fRTLRes,                      // Resolution of the RTL Bitmaps
    fEncoding,                    // RTL Image encoding
    fRowCount,                    // Number of pixel rows in the bitmap
    fImgBytes,                    // Bytes per row in the bitmap
    fRTLIndex,                    // Number of entries in the RTL palette
    fRotation,                    // Angle of the rotation
    fDataDone,                    // Number of bytes read from the stream
    fLastByte,                    // Position of the last read byte in a bitmap
    fStartX,
    fStartY,
    fLastX,
    fLastY,
    fImgCount,
    fHpError,
    fIncrement,                   // Defines call frequency of the OnProgress
    fIncSize,
    fPen        : integer;        // Current pen number
    fDoZoom,                      // Enable or disable HP zoom function
    fLoading,                     // Element in a loading stage;
    fBorder,
    fCanDo,                       // Flag of the major error
    fPCLFrame,                    // On if page size was entered through the PCL
    fPolyMode,                    // Polygone mode flag
    fRelative,                    // Absolute mode flag
    fRTLMode,                     // RTL vs HPGL flag
    fRasterMode,                  // Flag of the RTL Raster Mode
    fAborted,                     // User abort flag
    fScaling,                     // Scaling on flag
    fChangeSize,
    fPlotReady,                   // Flag indicating that view is ready
    fReading,                     // Job in progress flag
    fDrawing,                     // Drawing in progress flag
    fPenDown,                     // Do you need to ask?
    fGotScale,                    // Flag of the SC command
    fGotPoints,                   // Flag of the IP & IR commands
    fAutoPaint,                   // Draw plot immediately after reading
    fDoImage,
    fUseLast,
    fWinMode,
    fStarted,
    fMouseDn,
    fIgnorePs,                    // Page size defined in RTL, ignore PS cmd
    fDoViewport,
    fSizing     : boolean;        // On during form resizing
    fOnError    : THPGLError;     // Call back procedure for error report
    fOnProgress : THPGLProgr;     // File reading progress report procedure
    fStreamEnd,
    fStreamPtr  : TCharPtr;
    fElements   : TElementList;   // List of the HPGL Elements
    fLastCmd,
    fRTLCmd,
    fHPGL       : string;         // HPGL command string
    fParseBuff  : array[1..hpMaxParams] of string;  // Parsing buffer
    fDataBuff   : array[1..hpMaxParams] of real;    // Numerical parsing buffer
    fCmdTail,                                     // HPGL command parameters
    fPicName    : string;                         // Plot name from the file
    fFill       : THpFillInfo;
    fHgSeed,
    fHgImage    : HGLOBAL;
    fSeedRow,                                     // Image seed row buffer
    fImgBuffer  : PImageBuffer;
    fImage      : TImageData;
    fViewMode   : TViewMode;
    fImageParams : THPBmpData;
    {}
    procedure ResetData(DoAll : boolean);
    procedure SetFileName(const Value: string);
    procedure AddPoint(P : TPoint);
    procedure ClearList(Sw : integer);
    function  EOS : boolean;
    function  GetHPString : string;
    function  GetSpecHPString(Term: Char): string;
    procedure GetRTLString; //: string;
    function  GetChar: Char;
    function  PeekChar: Char;
    function  ProcessHPGL : integer;
    function  ProcessRTL  : integer;
    function  ProcessPJL  : integer;
    function  ParseCommand(Command : string;
                           Convert : boolean) : integer;
    function  FormatString(St : string) : string;
    procedure BeginPlot;
    procedure ShiftAndScale;
    procedure InputPoints;
    procedure InputPointsRelative;    
    procedure GetScales;
    procedure AddLine;
    procedure AddArc(ArcAbs : boolean);
    procedure SetBorder(const Value: boolean);
    procedure SetFrame(const Value: TColor);
    procedure SetPaperColor(const Value: TColor);
    procedure SetLineType;
    procedure GetPenWidth;
    procedure ReportError(Code : integer; Msg : string);
    procedure DecodePolyline;
    procedure AddWedge(Filled: boolean);
    procedure AddRectangle(Filled, Relative: boolean);
    procedure AddCircle;
    procedure SetRasterFill;
    procedure SetFillType;
    procedure SetPenColor;
    procedure SetScreenType;
    procedure SetAutoPaint(const Value: boolean);
    procedure AddWindow;
    function  ScalePoint(P : TPoint) : TPoint;
    procedure UpdatePenParams(N: integer);
    procedure SetAborted(const Value: boolean);
    procedure SetIncrement(const Value: integer);
    function GetExtents: TPoint;
    procedure SetDoZoom(const Value: boolean);
    procedure SetViewMode(const Value : TViewMode);
    {}
  protected
    procedure WM_Sizing(var Msg : TMessage); message WM_ENTERSIZEMOVE;
    procedure WM_Size(var Msg : TMessage); message WM_EXITSIZEMOVE;
    procedure WM_SysCom(var Msg : TMessage); message WM_SYSCOMMAND;
    property  Elements : TElementList read fElements;
    procedure Resize; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
                        X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
                      X, Y: Integer); override;
    {}
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Read;
    procedure   Paint; override;
    function    Extents : TRect;
    function    ElementType(Index : integer) : TElemType;
    function    ElementCount : integer;
    function    HPErrorMsg(Code: integer): string;
    procedure   Print;
    {}
    property    UsedPens : TUseArray read fPenUsed;
    property    HpPens   : THpPenArray read fPens;
    property    Canvas;
    property    HPPage : TPoint read fHPPage;
    property    PJLPage : TPoint read fPJPage;
    property    RTLPage : TPoint read fRTPage;
    property    RealPage : TPoint read GetExtents;
    {}
  published
    // Set to true to draw a plot immediately after reading it into memory
    property Autopaint : boolean read fAutoPaint write SetAutoPaint;
    property Align;
    property Anchors;
    // Set to true to abort reading
    property Aborted : boolean read fAborted write SetAborted;
    // Draws border around the drawing area
    property Border : boolean read fBorder write SetBorder;
    // Color of the border
    property BorderColor : TColor read fFrame write SetFrame;
    property Cursor;
    property Constraints;
    property DataDone : integer read fDataDone;
    property DoZoom : boolean read fDoZoom write SetDoZoom;
    property DoViewport : boolean read fDoViewPort write fDoViewport;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FileName : string read fPlotFile write SetFileName;
    property Font;
    property Left;
    property Height;
    property Increment : integer read fIncrement write SetIncrement;
    property PaperColor : TColor read fPaper write SetPaperColor;
    property Parent;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PlotReady : boolean read fPlotReady;
    property PopupMenu;
    property ShowHint;
    property Top;
    property Visible;
    property ViewMode : TViewMode read fViewMode write SetViewMode;
    property Width;
    {}
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnError    : THPGLError read fOnError write fOnError;
    property OnProgress : THPGLProgr read fOnProgress write fOnProgress;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnPaint     : TNotifyEvent read fOnPaint write fOnPaint;
    property OnPlotReady : TNotifyEvent read fOnReady write fOnReady;
    property OnAfterPaint : TNotifyEvent read fAfterPaint write fAfterPaint;
    property OnStartDock;
    property OnStartDrag;
  end;

  procedure Register;
{}
{$R HPGL2.RES}
implementation
uses
  Forms, Math;
{ THPGL2 }
const
  DefPWidth = 4;
  DefWidth : real = 0.35;
  MaxImage = 28 * 1024;
  Floor1    = #191;        // Start of the terminating characters 8 bit
  Floor2    = #95;
  Top1      = #255;        // Max char value for 8 bit polyline decoding
  Top2      = #127;
  Base1     = 64;
  Base2     = 32;
  Shift1    = 6;           // Shift for 8 bit polyline decoding
  Shift2    = 5;           // Shift for 7 bit polyline decoding
  Sub       = 63;
  ArcAngle  = 5;           // Angle increment for drawing curves in poly mode
  LblTerminator : Char = #3;
  VCharS  = ' +-,.0123456789';
  VCharN : string[14] = '+-.0123456789';
  VCharA = [' ', '+', '-', ',', '.', '0', '1', '2', '3',
            '4', '5', '6', '7', '8', '9'];
  DefColors : array[0..7] of TColor = (clWhite, clBlack, clRed, clGreen,
                                       clYellow, clBlue, clPurple, clAqua);
  PJLs : array[1..5] of string[15] = ('PAPERWIDTH',
                                      'PAPERLENGTH',
                                      'RESOLUTION',
                                      'LANGUAGE',
                                      'EOJ');
  liNode = 1;
  liPoly = 2;
  liBoth = 3;
 {}
var
  TF : Text;
{------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Samples', [THPGL2]);
end;

{------------------------------------------------------}
// Note: this hash function is described in "The C Programming Language"
//  by Brian Kernighan and Donald Ritchie, Prentice Hall
function Hash(const S : string) : integer;
var
  I : integer;
begin
  Result := 0;
  for I := 1 to Length(S) do
    Result := (Result * 31) + Ord(S[I]);
end;
{-----------------------------------------------------}
function IsNumber(Ch : Char) : boolean;
begin
  Result := Pos(Ch, VCharN) > 0;
end;
{------------------------------------------------------}
function THPGL2.HPErrorMsg(Code : integer) : string;
var
  St : string;
  PC : PChar;
begin
  if (Code = errHpOk) or (Code = 0) then
    Exit;
  Pc := StrAlloc(127);
  if LoadString(HInstance, HPErrorBase + Code, Pc, 127) > 0 then
    Result := StrPas(Pc)
  else
    Result := 'Unknown Error ' + IntToStr(Code);
  StrDispose(Pc);
end;
{------------------------------------------------------}
// Return the number of elements (polygons, polylines, e.t.c. in the list
function THPGL2.ElementCount : integer;
begin
  if fPlotReady then
    Result := fElements.Count
  else
    Result := -1;
end;
{------------------------------------------------------}
procedure THPGL2.ReportError(Code : integer; Msg : string);
begin
  if Assigned(fOnError) then
    fOnError(Self, Code, Msg)
  else
    Raise Exception.Create(Msg);
end;
{------------------------------------------------------}
constructor THPGL2.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fLoading := True;
  fCanDo := True;
  fOnError := nil;
  fOnProgress := nil;
  fOnPaint := nil;
  fAfterPaint := nil;
  fOnReady := nil;
  Align := alNone; //alClient;
  Width := 100;
  Height := 66;
  Color := clWhite;
  fBorder := True;
  fFrame := clBlack;
  fPaper := clWhite;
  fIncrement := 10;
  fReading := False;
  fDrawing := False;
  fSizing := False;
  fMouseDn := False;
  fDoZoom := True;
  fDoViewport := True;
  fViewMode := vmZoom;
  try
    fNodeList := THpNodeList.Create;
  except
    on E : Exception do
      begin
        ReportError(errHpNList, E.Message);
        fCanDo := False;
        Exit;
      end;
  end;
  try
    fPolyList := THpNodeList.Create;
  except
    on E : Exception do
      begin
        ReportError(errHpNList, E.Message);
        fCanDo := False;
        Exit;
      end;
  end;
  try
    fWinList := THpNodeList.Create;
  except
    on E : Exception do
      begin
        ReportError(errHpNList, E.Message);
//        fDoWin := False;
        Exit;
      end;
  end;
  fPlotFile := '';
  fReading := false;
  try
    fStream := TMemoryStream.Create;
  except
    on E : Exception do
      begin
        ReportError(errHpMStrm, E.Message);
        fCanDo := False;
        Exit;
      end;
  end;
  try
    fElements := TElementList.Create;
  except
    on E : Exception do
      begin
        ReportError(errHpEList, E.Message);
        fCanDo := False;
        Exit;
      end;
  end;
  fElements.Canvas := Canvas;
  fElements.BackColor := fPaper;
  fFill.Color := clWhite;
  fFill.Pattern := ftSolid;
  fFill.Data := 100;
  fElements.CurrFill := fFill;
  fElements.CurrPen := clBlack;
  fElements.PenStyle := psSolid;
  ResetData(True);
  fLoading := False;
end;
{------------------------------------------------------}
destructor THPGL2.Destroy;
begin
  ClearList(liBoth);
  if Assigned(fNodeList) then
    fNodeList.Free;
  if Assigned(fPolyList) then
    fPolyList.Free;
  if Assigned(fElements) then
    fElements.Free;
  if Assigned(fStream) then
    fStream.Free;
  inherited Destroy;
end;
{------------------------------------------------------}
procedure THPGL2.Resize;
begin
  inherited;
  if (csLoading in ComponentState) or fLoading or (not fDoViewport) then
    Exit;
  fElements.ViewPort := Rect(0, 0, Self.Width, Self.Height);
  if fPlotReady then
    Invalidate;
end;
{------------------------------------------------------}
procedure THPGL2.WM_Size(var Msg: TMessage);
begin
  inherited;
  fSizing := False;
  fElements.ViewPort := Rect(0, 0, Self.Width, Self.Height);
  if fPlotReady then
    Invalidate;
end;
{------------------------------------------------------}
procedure THPGL2.WM_Sizing(var Msg: TMessage);
begin
  fSizing := True;
end;
{------------------------------------------------------}
procedure THPGL2.WM_SysCom(var Msg: TMessage);
begin
  inherited;
  if (Msg.WParam = SC_MAXIMIZE) or (Msg.WParam = SC_RESTORE) then
    begin
      fElements.ViewPort := Rect(0, 0, Width, Height);
      if fPlotReady then
        Invalidate;
    end;
end;
{------------------------------------------------------}
procedure THPGL2.ClearList(Sw : integer);
begin
  if (Sw <> liPoly) and Assigned(fNodeList) then
    fNodeList.Clear;
  if (Sw > liNode) and Assigned(fPolyList) then
    fPolyList.Clear;
end;
{------------------------------------------------------}
// Default HPGL/2 settings. Used for IN command
procedure THPGL2.ResetData(DoAll: boolean);
var
  I : integer;
begin
  fScaleX := 1.0;
  fScaleY := 1.0;
//  fShiftX := 0;
//  fShiftY := 0;
  fScaling := false;
  fRelative := true;
  fPolyMode := False;
  fNodeList.Clear;
  fPolyList.Clear;
  fRotation := 0;
  fPicName := '';
  fGotScale := False;
  fGotPoints := False;
  if DoAll then
    begin
      fHPPage := Point(0, 0);
      fPJPage := Point(0, 0);
      fRTPage := Point(0, 0);
      fPage.X := 48 * HPGLSteps;
      fPage.Y := 36 * HPGLSteps;
      fRealP1.X := 0;
      fRealP1.Y := 0;
      fRealP2.X := fPage.X;
      fRealP1.Y := fPage.Y;
      fUserP1 := fRealP1;
      fUserP2 := fRealP2;
      fPenDown := False;
      fRotation := 0;
      fPenPos := Point(0, 0);
      fLastPos := fPenPos;
      fUseLast := False;
      fRtlCap := fPenPos;
      for I := 0 to hpMaxPens do
        begin
          if I <= 7 then
            fPens[I].Color := DefColors[I]
          else
            fPens[I].Color := 0;
          fPens[I].Width := -1.0;
          fPenUsed[I] := False;
        end;
      fPen := 0;
    end;
end;
{------------------------------------------------------}
// End Of Stream function
function  THPGL2.EOS : boolean;
begin
  Result := fDataDone >= fStream.Size;
end;
{-----------------------------------------------------}
// Define the scale to convert from user units to the HPGL units
procedure THPGL2.ShiftAndScale;
begin
  fScaleX := (fRealP2.X - fRealP1.X) / (fUserP2.X - fUserP1.X);
  fScaleY := (fRealP2.Y - fRealP1.Y) / (fUserP2.Y - fUserP1.Y);
end;
{-----------------------------------------------------}
// Get the HPGL/2 command from a stream
function THPGL2.GetHPString : string;
var
  P   : TCharPtr;
  Cnt : integer;
begin
  P := fStreamPtr;
  Cnt := 0;
  while (Pos(fStreamPtr^, VCharS) > 0) and (fDataDone < fStream.Size) do
    begin
      Inc(fStreamPtr);
      Inc(Cnt);
      Inc(fDataDone);
    end;
  if Cnt = 0 then
    begin
      Result := '';
      Exit;
    end;
  if fStreamPtr^ = '"' then
    begin
      repeat
        Inc(fStreamPtr);
        Inc(Cnt);
        Inc(fDataDone);
      until (fDataDone >= fStream.Size) or (fStreamPtr^ = '"');
      Inc(fStreamPtr);
      Inc(Cnt);
      Inc(fDataDone);
    end;
  SetLength(Result, Cnt);
  Move(P^, Result[1], Cnt);
  if fStreamPtr^ = ';' then
    begin
      Inc(fStreamPtr);
      Inc(fDataDone);
      Inc(Cnt);
    end;
  fStream.Position := fStream.Position + Cnt;
end;
{------------------------------------------------------}
// Get the HPGL command terminated by a particular character
function THPGL2.GetSpecHPString(Term : Char) : string;
var
  P   : TCharPtr;
  Cnt : integer;
begin
  P := fStreamPtr;
  Cnt := 0;
  while (fStreamPtr^ <> Term) and (fDataDone < fStream.Size) do
    begin
      Inc(fStreamPtr);
      Inc(Cnt);
      Inc(fDataDone);
    end;
  if Cnt = 0 then
    begin
      Result := '';
      Exit;
    end
  else if fStreamPtr^ = Term then
    begin
      Inc(fStreamPtr);
      Inc(fDataDone);
    end;
  SetLength(Result, Cnt);
  Move(P^, Result[1], Cnt);
  fStream.Position := fStream.Position + Cnt;
end;
{-----------------------------------------------------}
// Read an RTL command
procedure THPGL2.GetRTLString;
var
  P   : TCharPtr;
  Cnt : integer;
begin
  P := fStreamPtr;
  Cnt := 1;
  while ((fStreamPtr^ < 'A') or (fStreamPtr^ > 'Z')) and
         (fDataDone < fStream.Size) do
    begin
      Inc(fStreamPtr);
      Inc(Cnt);
      Inc(fDataDone);
    end;
  try
    SetLength(fRTLCmd, Cnt);
    Move(P^, fRTLCmd[1], Cnt);
  except
    on E : Exception do
      ShowMessage('Exception : ' + E.Message);
//    end;
  end;
  fStream.Position := fStream.Position + Cnt;
  Inc(fStreamPtr);
end;
{-----------------------------------------------------}
// Get one single character from a stream
function THPGL2.GetChar : Char;
begin
  if fDataDone < fStream.Size then
    begin
      Result := fStreamPtr^;
      Inc(fStreamPtr);
      Inc(fDataDone);
      fStream.Position := fStream.Position + 1;
    end
  else
    Result := #255;
end;
{-----------------------------------------------------}
function THPGL2.PeekChar : Char;
begin
  if fDataDone < fStream.Size then
    Result := fStreamPtr^
  else
    Result := #255;
end;
{------------------------------------------------------}
function THPGL2.ScalePoint(P: TPoint): TPoint;
begin
  if fScaling then
    begin
      Result.X := Round(P.X * fScaleX);
      Result.Y := Round(P.Y * fScaleY);
    end
  else
    Result := P;
end;
{------------------------------------------------------}
// Add a point to a particular list: node, vertex or window
procedure THPGL2.AddPoint(P : TPoint);
var
  Pp : TPoint;
begin
  Pp := P;
  case fRotation of
    000 : Pp := P;
    090 : begin
            Pp.X := -P.Y;
            Pp.Y := P.X;
          end;
    180 : begin
            Pp.X := -P.X;
            Pp.Y := -P.Y;
          end;
    270 : begin
            Pp.X := P.Y;
            Pp.Y := -P.X;
          end;
  end;
  try
    if fWinMode then
      fWinList.Add(Pp)
    else if fPolyMode then
      fPolyList.Add(Pp)
    else
      fNodeList.Add(Pp);
  except
    on E : Exception do
      begin
        if Assigned(fOnError) then
          fOnError(Self, errHpAddPnt, E.Message)
        else
          Raise Exception.Create(E.Message)
      end;
  end;
end;
{------------------------------------------------------}
// Strip all unneeded characters from a HPGL command
function THPGL2.FormatString(St : string) : string;
var
  S    : string;
  I, L : integer;
begin
  L := Length(St);
  if L = 0 then
    begin
      Result := '';
      Exit;
    end;
  St := Trim(St);
  S := '';
  I := 1;
  while I <= L do
    begin
      if not (St[I] in [' ', ',']) then
        begin
          S := S + St[I];
          Inc(I);
        end
      else
        begin
          S := S + ',';
          Inc(I);
          while (St[I] in [' ', ',']) and (I <= L) do
            Inc(I);
        end;
    end;
  Result := S;
end;
{------------------------------------------------------}
// Split a HPGL command into components
function  THPGL2.ParseCommand(Command : string;
                              Convert : boolean) : integer;
var
  I, J,
  L : integer;
  S    : string;
begin
  J := 0;
  S := FormatString(Command);
  L := Length(S);
  Result := 0;
  if L = 0 then
    Exit;
  I := Pos(',', S);
  while I > 0 do
    begin
      Inc(J);
      fParseBuff[J] := Copy(S, 1, I - 1);
      Delete(S, 1, I);
      I := Pos(',', S);
    end;
  Inc(J);
  fParseBuff[J] := S;
  if Convert then
    for I := 1 to J do
      try
        fDataBuff[I] := StrToFloat(fParseBuff[I]);
      except
        fDataBuff[I] := 0.0;
      end;
  Result := J;
end;
{------------------------------------------------------}
{$I-}
// Read and translate given plot file
procedure THPGL2.Read;
var
  Ch  : Char;
  Prc : integer;
  Cr  : TCursor;
  Sz  : integer;
  Fl  : file;
begin
  if (fPlotFile = '') or (not FileExists(fPlotFile)) or (not fCanDo) then
    begin
      if fPlotFile = '' then
        if Assigned(fOnError) then
          fOnError(Self, errHpFName, '');
      if not FileExists(fPlotFile) then
        if Assigned(fOnError) then
          fOnError(Self, errNoFile, '');
      Exit;
    end;
  AssignFile(Fl, fPlotFile);
  Reset(Fl, 1);
  if IOResult = 0 then
    begin
      Sz := FileSize(Fl);
      CloseFile(Fl);
    end
  else
    Sz := 99;
  if Sz = 0 then
    begin
      if Assigned(fOnError) then
        begin
          fOnError(Self, errZeroSize, '');
          Exit;
        end
      else
        Raise Exception.Create('File size is 0 bytes!');
    end;
  fElements.Clear;
  fSizing := false;
  fChangeSize := false;
  fGotScale := False;
  fGotPoints := False;
  fHpError := errHpOk;
  fWinMode := False;
  fPen := 0;
  fStarted := True;
  fReading := true;
  fDrawing := False;
  fPlotReady := False;
  fHpError := errHpOk;
  fAborted := False;
  fStarted := True;
  fDataDone := 0;
  fHPGL := '';
  fLastCmd := '';
  Cr := Cursor;
  Cursor := crHourglass;
  fStream.LoadFromFile(fPlotFile);
  fStreamPtr := fStream.Memory;
  fStreamEnd := fStreamPtr;
  Inc(fStreamEnd, fStream.Size);
  if fIncrement = 0 then
    fIncrement := 10;
  fIncSize := Round(fStream.Size / 100 * fIncrement);
  if fIncSize = 0 then
    fIncSize := fStream.Size div 10;
  repeat
//    Application.ProcessMessages;
    Ch := GetChar;
    case Ch of
      #27 : try
              fHpError := ProcessRTL;
            except
              fHPError := errRTLError;
              Break;
            end;
      '@' : try
              fHpError := ProcessPJL;
            except
              fHPError := errPJLError;
              Break;
            end;
      'A'..
      'Z' : try //begin
              fLastCmd := fHPGL;
              fHPGL := Ch + GetChar;
              fHpError := ProcessHPGL;
            except
              fHpError := errHPError;
              Break;
            end;
      ';' : ;
    end;
    Application.ProcessMessages;
    if Assigned(fOnProgress) then
      begin
        Prc := Round(fDataDone / fStream.Size * 100);
        if (Prc > 0) and ((Prc mod fIncrement) = 0) then
          fOnProgress(Self, Prc);
      end;
  until fPlotReady or fAborted or (fDataDone >= fStream.Size) or
        (fHpError <> errHpOk);
  fReading := false;
  Cursor := Cr;
  if Assigned(fOnProgress) then                         // 9/5/00
    fOnProgress(Self, 100);                             // 9/5/00
  if fPlotReady and Assigned(fOnReady) then             // 9/5/00
    fOnReady(Self);                                     // 9/5/00
  if (not fPlotReady) and (fHpError = errHpOk) then     // 9/5/00
    begin
      fHpError := errIncomplete;                        // 9/5/00
      fPlotReady := True;                               // 9/5/00      
      if Assigned(fOnError) then                        // 9/5/00
        fOnError(Self, fHpError, '');                   // 9/5/00
      if Assigned(fOnReady) then                        // 9/5/00
        fOnReady(Self);                                 // 9/5/00
    end
  else if (fHpError <> errHpOk) and Assigned(fOnError) then
    begin
      if fHpError = errHpInvHpCmd then
        fOnError(Self, fHpError, ' ' + fHPGL + ', ' + fLastCmd)
      else
        fOnError(Self, fHpError, '');
      fElements.Clear;
      Exit;
    end;
  fStream.Clear;
  fElements.InitView(Rect(0, 0, Width, Height));
  fElements.ViewReady := fPlotReady;
  if fAutoPaint and fPlotReady then
    Invalidate;
end;
{$I+}
{------------------------------------------------------}
procedure THPGL2.SetFileName(const Value: string);
begin
  if (not fReading) and (FileExists(Value)) then
    begin
      fPlotFile := Value;
      if csDesigning in ComponentState then
        Canvas.TextOut(10, 15, fPlotFile)
      else
        begin
          if fElements.Count > 0 then
            fElements.Clear;
        end;
    end;
end;
{------------------------------------------------------}
// BP command
procedure THPGL2.BeginPlot;
var
  I : integer;
  St : string;
  Ch : char;
  EndP : boolean;
begin
  EndP := False;
  while not EndP do
    begin
      Ch := PeekChar;
      if (Ch = ' ') or (Ch = ',') then
        while PeekChar in [' ', ','] do
          GetChar;
      Ch := PeekChar;
      if not (Ch in ['0'..'9']) then
        Exit;
      Ch := GetChar;
      I := Ord(Ch) - Ord('0');
      if (I < 1) or (I > 5) then
        Exit;
      repeat
        Ch := GetChar;
      until (Ch <> ',') and (Ch <> ' ');
      case I of
        1 : begin  // Picture name
              if Ch <> '"' then
                while Ch <> '"' do
                  begin
                    Application.ProcessMessages;
                    Ch := GetChar;
                  end;
              repeat
                Ch := GetChar;
                if Ch <> '"' then
                  fPicName := fPicName + Ch;
                Application.ProcessMessages;
              until Ch = '"';
            end;
        2..
        5 : begin
              St := '';
              while IsNumber(PeekChar) do
                St := St + GetChar;
            end;
      end;
      Ch := PeekChar;
      EndP := (Ch = ';') or (Ch in ['A'..'Z']);
    end;
end;
{-----------------------------------------------------}
procedure THPGL2.UpdatePenParams(N : integer);
begin
  fElements.CurrPen := fPens[N].Color;
  if fPens[N].Width >= 0 then
    fElements.PenWidth := fPens[N].Width;
  fFill.Color := fPens[N].Color;
  fElements.CurrFill := fFill;
  fPenUsed[N] := True;
end;
{-----------------------------------------------------}
// PC command
procedure THPGL2.SetPenColor;
var
  P, N  : integer;
  R, G, B  : BYTE;
begin
  fCmdTail := GetHpString;
  N := ParseCommand(fCmdTail, true);
  if N < 4 then
    Exit;
  P := Round(fDataBuff[1]);
  if N = 1 then
    begin
      case P of
        0 : fPens[fPen].Color := clWhite;
        1 : fPens[fPen].Color := clBlack;
        2 : fPens[fPen].Color := clRed;
        3 : fPens[fPen].Color := clGreen;
        4 : fPens[fPen].Color := clYellow;
        5 : fPens[fPen].Color := clBlue;
        6 : fPens[fPen].Color := clFuchsia;
        7 : fPens[fPen].Color := clAqua;
      end;
    end
  else
    begin
      R := Round(fDataBuff[2]);
      G := Round(fDataBuff[3]);
      B := Round(fDataBuff[4]);
      fPens[P].Color := RGB(R, G, B);
      if P = fPen then
        begin
          fElements.CurrPen := fPens[P].Color;
          if fPens[P].Width >= 0 then
            fElements.PenWidth := fPens[P].Width;
          fFill.Color := fPens[P].Color;
          fElements.CurrFill := fFill;
          fPenUsed[P] := True;
        end;
    end;
end;
{-------------------------------------------------}
// LT command
procedure THPGL2.SetLineType;
var
  N : integer;
begin
  fCmdTail := GetHpString;
  N := ParseCommand(fCmdTail, true);
  if N = 0 then
    fElements.PenStyle := psSolid
  else
    begin
      N := Abs(Round(fDataBuff[1]));
      case N of
        0 : fElements.PenStyle := psSolid;
        1 : fElements.PenStyle := psDot;
        2 : fElements.PenStyle := psDash;
        3 : fElements.PenStyle := psDash;
        4 : fElements.PenStyle := psDashDot;
        5 : fElements.PenStyle := psDashDot;
        6 : fElements.PenStyle := psDashDotDot;
        7 : fElements.PenStyle := psDashDotDot;
      end;
    end;
end;
{------------------------------------------------------}
// IR command
procedure THPGL2.InputPointsRelative;
var
  I : integer;
  J, K : real;
begin
  fCmdTail := GetHpString;
  Exit;
  I := ParseCommand(fCmdTail, true);
  if (I = 1) or (I = 3) then
    Exit;
  case I of
    0 : begin
          fRealP1.X := 0;
          fRealP1.Y := 0;
          fRealP2.X := fPage.X;
          fRealP2.Y := fPage.Y;
          fUserP1 := fRealP1;
          fUserP2 := fRealP2;
        end;
    2 : begin
          J := fRealP2.X - fRealP1.X;
          K := fRealP2.Y - fRealP1.Y;
          fRealP1.X := fRealP1.X + Round(J * fDataBuff[1] / 100);
          fRealP1.Y := fRealP1.Y + Round(K * fDataBuff[2] / 100);
          fRealP2.X := fRealP1.X + J;
          fRealP2.Y := fRealP1.Y + K;
        end;
  else  begin
          J := fRealP2.X - fRealP1.X;
          K := fRealP2.Y - fRealP1.Y;
          fRealP1.X := fRealP1.X + Round(fDataBuff[1] * J / 100);
          fRealP1.Y := fRealP1.Y + Round(fDataBuff[2] * K / 100);
          fRealP2.X := fRealP2.X - Round(fDataBuff[3] * J / 100);
          fRealP2.Y := fRealP2.Y - Round(fDataBuff[4] * K / 100);
        end;
  end;
  if fGotScale or fScaling then
    begin
      fGotScale := False;
      ShiftAndScale;
    end
  else
    fGotPoints := (I > 0);
end;
{------------------------------------------------------}
// IP command
procedure THPGL2.InputPoints;
var
  I : integer;
  J, K : real;
begin
  fCmdTail := GetHpString;
  I := ParseCommand(fCmdTail, true);
  if (I = 1) or (I = 3) then
    Exit;
  case I of
    0 : begin
          fRealP1.X := 0;
          fRealP1.Y := 0;
          fRealP2.X := fPage.X;
          fRealP2.Y := fPage.Y;
          fUserP1 := fRealP1;
          fUserP2 := fRealP2;
        end;
    2 : begin
          J := fRealP2.X - fRealP1.X;
          K := fRealP2.Y - fRealP1.Y;
          fRealP1.X := fDataBuff[1];
          fRealP1.Y := fDataBuff[2];
          fRealP2.X := fRealP1.X + J;
          fRealP2.Y := fRealP1.Y + K;
        end;
  else  begin
          fRealP1.X := fDataBuff[1];
          fRealP1.Y := fDataBuff[2];
          fRealP2.X := fDataBuff[3];
          fRealP2.Y := fDataBuff[4];
        end;
  end;
  if fGotScale or fScaling then
    begin
      fGotScale := False;
      ShiftAndScale;
    end
  else
    fGotPoints := (I > 0);
end;
{------------------------------------------------------}
procedure THPGL2.GetScales;
var
  I, K : integer;
begin
  fCmdTail := GetHpString;
  I := ParseCommand(fCmdTail, True);
  if (I < 4) then
    begin
      fScaling := false;
      fGotScale := false;
      Exit;
    end;
  if I > 4 then
    K := Round(fDataBuff[5])
  else
    K := 0;
  if K < 2 then
    begin
      fUserP1.X := fDataBuff[1];
      fUserP1.Y := fDataBuff[3];
      fUserP2.X := fDataBuff[2];
      fUserP2.Y := fDataBuff[4];
    end
  else
    begin
      if (fDataBuff[2] = 1.0) and (fDataBuff[4] = 1.0) then
        begin
          fScaling := False;
          fGotScale := False;
          fGotPoints := False;
          fRealP1.X := fDataBuff[1];
          fRealP1.Y := fDataBuff[2];
          fScaleX := 1.0;
          fScaleY := 1.0;
//          fShiftX := 0;
//          fShiftY := 0;
          Exit;
        end
      else
        begin
          fUserP1.X := fDataBuff[1];
          fUserP1.Y := fDataBuff[3];
          fUserP2.X := fUserP1.X + 1;
          fUserP2.Y := fUserP1.Y + 1;
          fRealP1.X := 0.0;
          fRealP1.Y := 0.0;
          fRealP2.X := fDataBuff[2];
          fRealP2.Y := fDataBuff[4];
        end;
    end;
  if fGotPoints or fScaling then
    begin
      ShiftAndScale;
      fGotPoints := false;
    end
  else
    fGotScale := True;
  fScaling := true;
end;
{------------------------------------------------------}
procedure THPGL2.AddWindow;
var
  N : integer;
  P1, P2 : TPoint;
begin
  fCmdTail := GetHpString;
  N := ParseCommand(fCmdTail, True);
  if N = 0 then
    begin
      fElements.Add(etFullWin, nil);
    end;
  if N = 4 then
    begin
      fWinMode := True;
      P1 := ScalePoint(Point(Round(fDataBuff[1]), Round(fDataBuff[2])));
      P2 := ScalePoint(Point(Round(fDataBuff[3]), Round(fDataBuff[4])));
      if P1.X > P2.X then
        begin
          N := P1.X;
          P1.X := P2.X;
          P2.X := N;
        end;
      if P1.Y > P2.Y then
        begin
          N := P1.Y;
          P1.Y := P2.Y;
          P2.Y := N;
        end;
      AddPoint(P1);
      AddPoint(P2);
      fElements.Add(etWin, fWinList);
      fWinList.Clear;
      fWinMode := False;
    end;
end;
{-------------------------------------------------}
procedure THPGL2.AddLine;
var
  St, S1 : string;
  I,
  X, Y   : integer;
  Add    : boolean;
begin
  St := FormatString(GetHPString);
  if St = '' then
    Exit;
  Add := false;
  repeat
    I := Pos(',', St);
    if I > 0 then
      begin
        S1 := Copy(St, 1, I - 1);
        Delete(St, 1, I);
        if Add then
          begin
            try
              Y := StrToInt(S1);
            except
              Y := 0;
            end;
            if fScaling then
              Y := Round(Y * fScaleY); // + fShiftY;
            if fRelative then
              fPenPos.Y := fPenPos.Y + Y
            else
              fPenPos.Y := Y;
            if fPenDown then
              begin
                if fUseLast then                        // 9/5/00
                  AddPoint(fLastPos);                   // 9/5/00
                fUseLast := False;                      // 9/5/00
                AddPoint(fPenPos);
              end
            else
              begin
                fUseLast := True;                       // 9/5/00
                fLastPos := fPenPos;                    // 9/5/00
              end;
          end
        else
          begin
            try
              X := StrToInt(S1);
            except
              X := 0;
            end;
            if fScaling then
              X := Round(X * fScaleY); // + fShiftX;
            if fRelative then
              fPenPos.X := fPenPos.X + X
            else
              fPenPos.X := X;
          end;
        Add := not Add;
      end;
    until I = 0;
  if (Length(St) > 0) and Add then
    begin
      try
        Y := StrToInt(S1);
      except
        Y := 0;
      end;
      if fScaling then
        Y := Round(Y * fScaleY); // + fShiftY;
      if fRelative then
        fPenPos.Y := fPenPos.Y + Y
      else
        fPenPos.Y := Y;
      if fPenDown then
        begin
          if fUseLast then                              // 9/5/00
            AddPoint(fLastPos);                         // 9/5/00
          fUseLast := False;                            // 9/5/00
          AddPoint(fPenPos);
        end
      else
        begin
          fUseLast := True;                             // 9/5/00
          fLastPos := fPenPos;                          // 9/5/00
        end;
    end;
  if (not fPolyMode) and fPenDown then
    begin
      if fNodeList.Count > 1 then
        fElements.Add(etPline, fNodeList);
      fNodeList.Clear;
    end;
end;
{------------------------------------------------------}
procedure THPGL2.AddArc(ArcAbs : boolean);
var
  I, R,
  X, Y,
  X1, Y1 : integer;
  Alpha,
  Beta   : double;
  S1, S2 : extended;
begin
  fCmdTail := GetHPString;
  I := ParseCommand(fCmdTail, True);
  if I < 3 then
    Exit;
  // Get coords of the center of an arc
  if fScaling then
    begin
      fDataBuff[1] := fDataBuff[1] * fScaleX;
      fDataBuff[2] := fDataBuff[2] * fScaleY;
    end;
  if ArcAbs then
    begin
      X := Round(fDataBuff[1]);
      Y := Round(fDataBuff[2]);
    end
  else
    begin
      X := fPenPos.X + Round(fDataBuff[1]);
      Y := fPenPos.Y + Round(fDataBuff[2]);
    end;
  // Arc's sweep angle
  Beta := fDataBuff[3];
  if ((fPenPos.X = X) and (fPenPos.Y = Y)) or (Beta = 0) then
    Exit;
  // If a sweep angle is greater than 360 - adjust it to 0 - 359;
  if Beta > 360 then
    Beta := Round(Beta) mod 360;
  S1 := Sqr(fPenPos.X - X);
  S2 := Sqr(fPenPos.Y - Y);
//  S1 := S1 * S1;
//  S2 := S2 * S2;
  // Arc's radius
  R := Round(Sqrt(S1 + S2));
  if R = 0 then
    Exit;
  // Angle between the radius and X axis in degrees
  Alpha := ArcSin((fPenPos.Y - Y) / R) * 180 / Pi;
  // ArcSin returns result in the -Pi / 2 to Pi / 2 range (270 to 90 degrees)
  // If pen X is less than center X we are in a quadrant III or IV
  if fPenPos.X < X then
    begin
  // Adjusted angle in the quadrants II
      if fPenPos.Y >= Y then
        Alpha := 180 - Alpha
      else
  // Adjusted angle in the quadrants III      
        Alpha := 270 + Alpha;
    end;
  S1 := Alpha + Beta;
  if S1 > 360 then
    while S1 > 360 do
      S1 := S1 - 360
  else if S1 < 0 then
    S1 := 360 + S1;
  // End coordinates of an arc
  X1 := X + Round(R * Cos(S1 * Pi / 180));
  Y1 := Y + Round(R * Sin(S1 * Pi / 180));
  // Arc defined in the clockwise direction. End points will be swapped
  if Beta < 0 then
    begin
      I := fPenPos.X;
      fPenPos.X := X1;
      X1 := I;
      I := fPenPos.Y;
      fPenPos.Y := Y1;
      Y1 := I;
    end;
// In polygon mode arc will be represented as series of points
  if fPolyMode then
    begin
      Alpha := Alpha * Pi / 180;
      Beta := Alpha + (Beta * Pi / 180);
      AddPoint(fPenPos);
      while Alpha <= Beta do
        begin
          Alpha := Alpha + Delta;
          fPenPos.X := Round(R * Cos(Alpha));
          fPenPos.Y := Round(R * Sin(Alpha));
          AddPoint(fPenPos);
        end;
    end
  else
    begin
      AddPoint(Point(X - R, Y - R));  // Top-Left of the arc box
      AddPoint(Point(X + R, Y + R));  // Bottom-Right of the box
      AddPoint(fPenPos);              // Start point
      AddPoint(Point(X1, Y1));        // End point
      fElements.Add(etArc, fNodeList);
    end;
  fPenPos := Point(X1, Y1);
end;
{------------------------------------------------------}
procedure THPGL2.AddWedge(Filled : boolean);
var
  I, R : integer;
  Alpha,
  Beta  : double;
begin
  fPolyList.Clear;
  fCmdTail := GetHPString;
  I := ParseCommand(fCmdTail, True);
  if I < 3 then
    Exit;
  // Get wedge radius
  R := Round(fDataBuff[1]);
  if R = 0 then
    Exit;
  Alpha := fDataBuff[2];
  Beta := fDataBuff[3];
  if R < 0 then
    Exit;
end;
{------------------------------------------------------}
procedure THPGL2.AddRectangle(Filled, Relative : boolean);
var
  I, X, Y : integer;
  SaveP   : boolean;
begin
  fPolyList.Clear;
  SaveP := fPolyMode;
  fPolyMode := True;
  fCmdTail := GetHPString;
  I := ParseCommand(fCmdTail, True);
  if I < 2 then
    Exit;
  if fScaling then
    begin
      X := Round(fDataBuff[1] * fScaleX);
      Y := Round(fDataBuff[2] * fScaleY);
    end
  else
    begin
      X := Round(fDataBuff[1]);
      Y := Round(fDataBuff[2]);
    end;
  if Relative then
    begin
      X := fPenPos.X + X;
      Y := fPenPos.Y + Y;
    end;
  AddPoint(fPenPos);
  AddPoint(Point(fPenPos.X, Y));
  AddPoint(Point(X, Y));
  AddPoint(Point(X, fPenPos.Y));
  if Filled then
    fElements.Add(etPgon, fPolyList)
  else
    fElements.Add(etPline, fPolyList);
  fPolyMode := SaveP;
end;
{-------------------------------------------------}
procedure THPGL2.AddCircle;
var
  I, Rx,
  Ry, X  : integer;
  Alpha  : real;
begin
  fCmdTail := GetHPString;
  I := ParseCommand(fCmdTail, True);
  if I = 0 then
    Exit;
  if fScaling then
    begin
      Rx := Round(fDataBuff[1] * fScaleX); // + fShiftX;
      Ry := Round(fDataBuff[1] * fScaleY); // + fShiftY;
    end
  else
    begin
      Rx := Round(fDataBuff[1]);
      Ry := Round(fDataBuff[1]);
    end;
  if fPolyMode then
    begin
      I := 0;
      Alpha := 0.0;
      X := fPenPos.X + Rx;
      Ry := fPenPos.Y;
      repeat
        AddPoint(Point(X, Ry));
        Alpha := Alpha + Delta;
        X := fPenPos.X + Round(Rx * Cos(Alpha));
        Ry := fPenPos.Y + Round(Rx * Sin(Alpha));
      until I >= C_To_R;
    end
  else
    begin
      AddPoint(Point(fPenPos.X - Rx, fPenPos.Y - Ry));
      AddPoint(Point(fPenPos.X + Rx, fPenPos.Y + Ry));
      fElements.Add(etCircle, fNodeList);
    end;
end;
{------------------------------------------------------}
procedure THPGL2.SetRasterFill;
begin
  fCmdTail := FormatString(GetHPString);
  if Length(fCmdTail) = 0 then
    Exit;
end;
{------------------------------------------------------}
procedure THPGL2.SetFillType;
var
  I, K : integer;
begin
  fCmdTail := GetHPString;
  I := ParseCommand(fCmdTail, True);
  if I > 0 then
    K := Round(fDataBuff[1])
  else
    K := 1;
  case K of
    1, 2 : fFill.Pattern := ftSolid;
    3    : fFill.Pattern := ftLines;
    4    : fFill.Pattern := ftCross;
    10   : begin
             fFill.Pattern := ftShade;
             fFill.Data := Round(fDataBuff[2]);
           end;
    11   : begin
             fFill.Pattern := ftBitmap;
             fFill.Data := Round(fDataBuff[1]);
           end;
  end;
  fElements.CurrFill := fFill;
end;
{------------------------------------------------------}
procedure THPGL2.GetPenWidth;
var
  N : integer;
begin
  fCmdTail := GetHPString;
  N := ParseCommand(fCmdTail, True);
  if (N = 0) then
    Exit;
  fElements.PenWidth := fDataBuff[1];
  if (N > 1) then
    begin
      N := Round(fDataBuff[2]);
      if N > 0 then
        fPens[N].Width := fDataBuff[1];
    end;
end;
{------------------------------------------------------}
procedure THPGL2.SetScreenType;
var
  I, N : integer;
begin
  fCmdTail := GetHPString;
  N := ParseCommand(fCmdTail, True);
  if (N = 0) then
    begin

      Exit;
    end;
  I := Round(fDataBuff[1]);
  if I >= 1 then
    N := Round(fDataBuff[2]);
  case I of
    0 : fFill.Pattern := ftSolid;
    1 : fFill.Pattern := ftShade;
    2 : ;
    21 : begin
           if N <= 4 then
             begin
               fFill.Pattern := ftLines;
               case N of
                 1 : fFill.Data := 0;
                 2 : fFill.Data := 90;
                 3 : fFill.Data := 45;
                 4 : fFill.Data := 135;
               end
             end
           else
             begin
               fFill.Pattern := ftCross;
               if N = 5 then
                 fFill.Data := 0
               else
                 fFill.Data := 45;
             end;
         end;
  end;
  fElements.CurrFill := fFill;
end;
{------------------------------------------------------}
procedure THPGL2.DecodePolyline;
var
  EndP,
  PUp, Rel  : boolean;
  X, Y      : integer;
  P         : TPoint;
  Flr, Ch   : char;
  Sft, Pen  : integer;
{-----------------------------}
  function CalcCoord(Second : boolean) : integer;
  var
    A, B, S : integer;
  begin
    S := 0;
    B := 0;
    if Second then
      Ch := GetChar;
    while Ch < Flr do
      begin
        A := Ord(Ch) - Sub;
        B := B + (A ShL S);
        S := S + Sft;
        Ch := GetChar;
      end;
    A := Ord(Ch) - Ord(Flr);
    if S > 0 then
      B := B + (A ShL S)
    else
      B := A;
    A := 1;
    if Odd(B) then
      begin
        A := -1;
        Dec(B);
      end;
    Result := (B div 2) * A;
  end;
{-----------------------------}
begin
  EndP := False;
  Flr := Floor1;
  Sft := Shift1;
  P := fPenPos;
  PUp := False;
  Rel := fRelative;
  fRelative := True;
  repeat
    Ch := GetChar;
//    if Flr = Floor2 then
//      Ch := Char(Ord(Ch) and $7F); 
    if Ch < #63 then  // The character is a flag
      case Ch of
        ':' : begin  // Change current pen outsite the polygon mode
                Pen := CalcCoord(True);
                if not fPolyMode then
                  if ((Pen >= 0) and (Pen < NumPens)) then
                    begin
                      fPen := Pen;
                      UpdatePenParams(Pen);
                    end;
              end;
        '<' : begin  // Raise the pen
                PUp := True;
              end;
        '>' : begin  // Fractional data, not implemented yet
                PUp := True;
              end;
        '=' : begin  // Absolute coordinates are on
                fRelative := False;
              end;
        '7' : begin  // 7 bit polyline encoding
                Flr := Floor2;
                Sft := Shift2;
              end;
        ';' : EndP := True;  // End of the PE command
        else begin  // Invalid character, skip to the nearest ;
               while (Ch <> ';') and (fDataDone < fStream.Size) and
                     (not fAborted) do
                 Ch := GetChar;
               if Assigned(fOnError) then
                 fOnError(Self, errHpInvPln, '');
               Exit;
             end;
      end
    else  // Process coordinates
      begin
        Application.ProcessMessages;
        X := CalcCoord(False);
        Y := CalcCoord(True);
        if fScaling then
          begin
            X := Round(X * fScaleX);
            Y := Round(Y * fScaleY);
          end;
        if fRelative then
          P := Point(P.X + X, P.Y + Y)
        else
          begin
            fRelative := True;
            P := Point(X, Y);
          end;
        if fPolyMode then
          AddPoint(P)
        else
          begin
            if PUp then
              begin
                // Pen is up - check if we have to add polyline to a list
                PUp := False;
                if fNodeList.Count > 1 then
                  fElements.Add(etPline, fNodeList);
                fNodeList.Clear;
                AddPoint(P);
              end
            else
              AddPoint(P);
          end;
      end;
  until EndP or fAborted or (fDataDone >= fStream.Size);
  fRelative := Rel;
  fPenPos := P;
  // Add a polyline if we are not in the polygon mode.
  if (fNodeList.Count > 1) and (not fPolyMode) then
    begin
      fElements.Add(etPline, fNodeList);
      fNodeList.Clear;
    end
  else if fAborted then
    fNodeList.Clear;
end;
{-------------------------------------------------}
function THPGL2.ProcessHPGL: integer;
var
  I, N, Sw : integer;
begin
  Sw := Hash(fHPGL);
  Result := errHpOk;
  case Sw of
    2080 : begin   // Arc Absolute
             AddArc(True);
           end;
    2082 : begin  // Anchor Corner
             GetHpString;
           end;
    2083 : begin  // Alternate font Definition
             GetHpString;
           end;
    2097 : begin  // Arc Relative
             AddArc(False);
           end;
    2099 : begin  // Arc absolute Three points
             GetHpString;
           end;
    2126 : begin  // Begin Pplot
             ResetData(true);
             BeginPlot;
           end;
    2128 : begin  // Bezier Relative
             GetHpString;
           end;
    2136 : begin  // BeZier absolute
             GetHpString;
           end;
    2147 : begin  // Character Fill mode
             GetHpString;
           end;
    2150 : begin  // CIrcle
             AddCircle;
           end;
    2156 : begin  // COmment
             GetHpString;
           end;
    2157 : begin  // Character Plot
             GetHpString;
           end;
    2159 : begin  // Color Range
             GetHpString;
           end;
    2161 : begin  // Chord Tolerance
             GetHpString;
           end;
    2175 : begin  // Digitize Clear
           end;
    2178 : begin  // DeFault values
             ResetData(false);
           end;
    2181 : begin  // absolute DIrection
             GetHpString;
           end;
    2184 : begin  // DownLoad character
             GetHpString;
           end;
    2188 : begin  // Digitize Point
           end;
    2190 : begin  // Direction Relative
             GetHpString;
           end;
    2192 : begin  // Define label Terminator
             fCmdTail := FormatString(GetHPString);
             LblTerminator := fCmdTail[1];
           end;
    2194 : begin  // Define Variable text path
             GetHpString;
           end;
    2204 : begin  // Edge rectangle Absolute
             AddRectangle(False, False);
           end;
    2206 : begin  // Enable Cutter
             GetHpString;
           end;
    2219 : begin  // Edge Polygone
             if fPolyList.Count > 1 then
               fElements.Add(etPline, fPolyList);
           end;
    2221 : begin  // Edge rectangle Relative
             AddRectangle(False, True);
           end;
    2222 : begin  // Extra Space
             GetHpString;
           end;
    2226 : begin  // Edge Wedge
             AddWedge(False);
           end;
    2243 : begin  // primary Font by Id
             GetHpString;
           end;
    2248 : begin  // secondary Font by Name
             GetHpString;
           end;
    2250 : begin  // Fill Polygone
             if fPolyList.Count > 1 then
               fElements.Add(etPgon, fPolyList);
           end;
    2252 : begin  // FRame advance
           end;
    2254 : begin  // Fill Type
             SetFillType;
           end;
    2341 : begin  // INitialize
             ResetData(true);
           end;
    2343 : begin  // Input Points
             InputPoints;
           end;
    2345 : begin  // Input points Relative
             InputPointsRelative;
           end;
    2350 : begin  // Input Window
             AddWindow;
           end;
    2421 : begin  // Line Attributes
             GetHpString;
           end;
    2422 : begin  // LaBel
             GetSpecHpString(LblTerminator);
           end;
    2433 : begin  // Label Mode
             GetHpString;
           end;
    2435 : begin  // Label Origin
             GetHpString;
           end;
    2440 : begin  // Line Type
             SetLineType;
           end;
    2454 : begin  // Merge Control
             fCmdTail := GetHPString;
             N := ParseCommand(fCmdTail, True);
             if (N = 0) or (fDataBuff[1] = 0) then
               Elements.Merge := False
             else
               Elements.Merge := True;
           end;
    2458 : begin  // MessaGe
             GetHpString;
           end;
    2471 : begin  // Media Type
             GetHpString;
           end;
    2498 : begin  // Number of Pens
             fCmdTail := GetHPString;
             I := ParseCommand(fCmdTail, True);
             if I = 0 then
               fPenCount := 8
             else
               begin
                 fPenCount := Round(fDataBuff[1]);
                 if fPenCount = 0 then
                   fPenCount := 256;
              end;
           end;
    2500 : begin  // Not Ready
             GetHpString;
           end;
    2517 : begin  // Output Digitized point
           end;
    2518 : begin  // Output Error
           end;
    2521 : begin  // Output Hard clip limits
           end;
    2522 : begin  // Output Identification
           end;
    2529 : begin  // Output P1 & p2
           end;
    2532 : begin  // Output Status
           end;
    2545 : begin  // Plot Absolute
             fRelative := false;
             AddLine;
           end;
    2547 : begin  // Pen Color
             SetPenColor;
           end;
    2548 : begin  // Pen Down
             fPenDown := true;
             AddLine;
           end;
    2549 : begin  // Polyline Encode
             DecodePolyline;
           end;
    2551 : begin  // advance full PaGe
             GetHpString;
             fPlotReady := True;
           end;
    2557 : begin  // Polygone Mode
             fCmdTail := GetHPString;
             N := ParseCommand(fCmdTail, True);
             if N = 0 then
               I := 0
             else
               I := Round(fDataBuff[1]);
             case I of
               0 : begin
                     fPolyMode := True;
                     fPolyList.Clear;
                   end;
               1 : fPolyMode := True;
               2 : fPolyMode := False;
              end;
           end;
    2560 : begin  // Pixel Placement
             GetHpString;
           end;
    2562 : begin  // Plot Relative
             fRelative := true;
             AddLine;
           end;
    2563 : begin  // Plot Size
             fCmdTail := GetHPString;
             N := ParseCommand(fCmdTail, True);
             if N = 2 then
               begin
                 fPage.X := Round(fDataBuff[1] / HPGLSteps);
                 fPage.Y := Round(fDataBuff[2] / HPGLSteps);
                 fHpPage := fPage;
                 if fPage.Y > fPage.X then
                   fRotation := 90;
               end;
           end;
    2565 : begin  // Pen Up
             fPenDown := false;
             AddLine;
           end;
    2567 : begin  // Pen Width
             GetPenWidth;
           end;
    2587 : begin  // QuaLity
             GetHpString;
           end;
    2607 : begin  // fill Rectangle Absolute
             AddRectangle(True, False);
           end;
    2612 : begin  // Raster Fill definition
             SetRasterFill;
           end;
    2621 : begin  // ROtate
             fCmdTail := GetHPString;
             N := ParseCommand(fCmdTail, True);
             if N > 0 then
               N := Round(fDataBuff[1])
             else
               N := 0;
             if fPage.Y < fPage.X then
               fRotation := N
             else
               fRotation := fRotation + N;
           end;
    2622 : begin  // RePlot
             GetHpString;
           end;
    2624 : begin  // fill Rectangle Relative
             AddRectangle(True, True);
           end;
    2626 : begin  // Relative arc Three points
           end;
    2638 : begin  // Select Alternative font
             GetHpString;
           end;
    2639 : begin  // ScalaBle font
             GetHpString;
           end;
    2640 : begin  // SCale
             GetScales;
           end;
    2641 : begin  // StanDard font definition
             GetHpString;
           end;
    2646 : begin  // absolute character SIze
             GetHpString;
           end;
    2649 : begin  // character SLant
             GetHpString;
           end;
    2650 : begin  // Symbol Mode
             GetHpString;
           end;
    2653 : begin  // Select Pen
             fCmdTail := GetHPString;
             N := ParseCommand(fCmdTail, True);
             if N > 0 then
               begin
                 N := Round(fDataBuff[1]);
                 fPen := N;
                 if N > 0 then
                   UpdatePenParams(N)
                 else
                   fPlotReady := True;
               end
             else
               fPlotReady := True;
           end;
    2655 : begin  // character Size Relative
             GetHpString;
           end;
    2656 : begin  // Select Standard font
           end;
    2657 : begin  // SorT
             GetHpString;
           end;
    2659 : begin  // Screened Vectors
             SetScreenType;
           end;
    2672 : begin  // Transparent Data
             GetHpString;
           end;
    2686 : begin  // TRansparncy mode
             fCmdTail := GetHPString;
             I := ParseCommand(fCmdTail, True);
             if (I = 0) or (fDataBuff[1] = 0) then
               fElements.TransMode := False
             else
               fElements.TransMode := True;
           end;
    2711 : begin  // User defined Line type
             GetHpString;
           end;
    2749 : begin  // Velosity Select
             GetHpString;
           end;
    2768 : begin  // fill WedGe
             AddWedge(True);
           end;
    2782 : begin  // pen Width Units
             GetHpString;
           end;
    else   Result := errHpInvHpCmd;
  end;
end;
{------------------------------------------------------}
function THPGL2.ProcessPJL: integer;
var
  St,
  St1 : string;
  Ch  : char;
  I,
  J   : integer;
begin
  Result := errHpInvPjCmd;
  St := '';
  for I := 1 to 3 do
    St := St + GetChar;
  if UpperCase(St) <> 'PJL' then
    Exit;
  Result := errHpOk;
  St := '';
  Ch := #0;
  while (Ch <> #13) and (Ch <>  #10) and (fDataDone < fStream.Size)do
    begin
      Ch := GetChar;
      if Ch >= #13 then
        St := St + Ch;
      Application.ProcessMessages;
    end;
  for I := 1 to 5 do
    if Pos(PJLs[I], St) > 0 then
      begin
        J := Pos('=', St);
        if J = 0 then
          Exit;
        Delete(St, 1, J);
        if St[1] = ' ' then
          while St[1] = ' ' do
            Delete(St, 1, 1);
        St1 := '';
        for J := 1 to Length(St) do
          if IsNumber(St[J]) then
            St1 := St1 + St[J];
        try
          J := StrToInt(St1);
        except
          Exit;
        end;
        case I of
          1 : begin
                fPage.X := Round(J * 1.411);
                if fPage.Y > fPage.X then
                  fRotation := 90
                else
                  fRotation := 0;
                fPJPage := fPage;
                Exit;
              end;
          2 : begin
                fPage.Y := Round(J * 1.411);
                if fPage.Y  > fPage.X then
                  fRotation := 90
                else
                  fRotation := 0;
                fPJPage := fPage;
                Exit;
              end;
          3 : begin
                fRTLRes := J;
                fRtlScale := HPGLSteps / fRtlRes;
                Exit;
              end;
          5 : fPlotReady := True;
        end;
      end;
end;
{-------------------------------------------------}
function THPGL2.ProcessRTL : integer;
var
  Ch, Ch1 : char;
  St1 : string[20];
  N, I, K  : integer;
  X, Y : integer;
  IBuf  : array[1..18] of byte;
  R, G, B : byte;
  {----------------------------------------}
  procedure GetImageMemory;
  begin
    fRasterMode := False;
    fSeedRow := nil;
    fImgCount := 0;
    if (fRtlWidth = 0) or (fRtlHeight = 0) then
      Exit;
    fOldBytes := fImgBytes;
    fOldCount := fImgCount;
    fDoImage := False;
    fImgBytes := fRtlWidth div 8;
    if fImgBytes = 0 then
      fImgBytes := 1
    else
      begin
        if (fRtlWidth mod 8) > 0 then
          Inc(fImgBytes);
      end;
    fImgCount := fImgBytes * fRtlHeight;
    if fImgCount = 0 then
      Exit;
    fHgImage := GlobalAlloc(GPTR, fImgCount);
    if fHgImage = 0 then
      Exit;
    fImgBuffer := GlobalLock(fHgImage);
    fHgSeed := GlobalAlloc(GPTR, fImgCount);
    if fHgSeed <> 0 then
      fSeedRow := GlobalLock(fHgSeed)
    else
      fSeedRow := nil;
    FillChar(fImgBuffer^[0], fImgCount, $00);
    if Assigned(fSeedRow) then
      FillChar(fSeedRow^[0], fImgBytes, $00);
    fRasterMode := True;
    fDoImage := True;
  end;
  {------------------------------------------}
  procedure FreeImageMemory;
  begin
    try
      if Assigned(fImgBuffer) then
        begin
          GlobalUnlock(fHgImage);
          GlobalFree(fHgImage);
        end;
      if Assigned(fSeedRow) then
        begin
          GlobalUnlock(fHgSeed);
          GlobalFree(fHgSeed);
        end;
    finally
      fImgBuffer := nil;
      fSeedRow := nil;
    end;
  end;
  {------------------------------------------}
  // We have to rotate an image due to the difference between HPGL and RTL
  // coordinate systems.
  procedure RotateImage;
  var
    WCol, WCnt, I, J, K, WBytes, Cnt : integer;
    Mask1, Mask2  : byte;
    fHgRImage : HGLOBAL;
  begin
    WBytes := fRtlHeight div 8;
    if (fRtlHeight mod 8) > 0 then
      Inc(WBytes);
    if Odd(WBytes) then
      Inc(WBytes);
    Cnt := WBytes * fImgBytes * 8;
    if Cnt = 0 then
      Exit;
    fHgRImage := GlobalAlloc(GPTR, Cnt);
    if fHgRImage = 0 then
      Exit;
    fImage.Image := GlobalLock(fHgRImage);
    if not Assigned(fImage.Image) then
      begin
        I := GetLastError;
        ShowMessage('Allocation error ' + IntToStr(I));
        GlobalFree(fHgRImage);
        Exit;
      end;
    Mask2 := $80;  // Mask for the destination image
    WCol := 0;
    FillChar(fImage.Image^[0], Cnt, $00);
    // For each row in the source image
    for I := 0 to fRtlHeight - 1 do
      begin
        // Calculate starting location in the destination image
        WCnt := WCol;
        // for each byte in the image row
        for J := fImgBytes - 1 downto 0 do
          begin
            K := I * fImgBytes + J;
            Mask1 := $01;  // Mask for the source image
            // Check each bit in the source byte. If bit is not set then
            // set appropriate bit in the destination image
            for Cnt := 1 to 8 do
              begin
                if (fImgBuffer^[K] and Mask1) = 0 then
                  fImage.Image^[WCnt] := fImage.Image^[WCnt] or Mask2;
                Mask1 := Mask1 ShL 1;  // Next bit
                // Next column byte in the destination image
                Inc(WCnt, WBytes);
              end;
          end;
        Mask2 := Mask2 ShR 1;  // Next bit
        if Mask2 = 0 then      // Next column of the destination image;
          begin
            Mask2 := $80;
            Inc(WCol);
          end;
      end;
    with fImage do
      begin
        Width := fRtlHeight;
        Height := fImgBytes * 8;
        Resolut  := fRtlRes;
        Bits := 1;
        Bytes := WBytes * Height;
        X := Round(fRTLCap.Y * fRTLScale);
//        Y := Round((fRTLCap.X  + fRtlWidth) * fRTLScale);
        Y := Round(fRTLCap.X  * fRTLScale);
      end;
    fElements.Add(etImage, @fImage);
    if not GlobalUnlock(fHgRImage) then
      begin
        if GetLastError = NO_ERROR then
          try
            GlobalFree(fHgRImage);
          except
            on E : Exception do
              ShowMessage(E.Message);
          end;
      end
    else
      GlobalFree(fHgRImage);
    fImage.Image := nil;
  end;
  {----------------------------------------}
  // Get TIFF Pack Bit encoded row of an image data
  procedure GetTIFFRow(Count : integer);
  var
    Cnt : ShortInt;
    Bt  : byte;
    I, J, P : integer;
  begin
    if Count = 0 then
      begin
        Move(fSeedRow^[0], fImgBuffer^[fRowCount * fImgBytes], fImgBytes);
        Exit;
      end;
    J := 0;
    P := fRowCount * fImgBytes;
    repeat
      Inc(J);
      Cnt := ShortInt(GetChar);
      if Cnt >= 0 then
        begin
          for I := 0 to Cnt do
            fImgBuffer^[P + Cnt] := byte(GetChar);
          P := P + Cnt + 1;
          Inc(J, Cnt + 1);
        end
      else
        begin
          while Cnt = - 128 do
            Cnt := byte(GetChar);
          Cnt := Abs(Cnt);
          Bt := byte(GetChar);
          for I := 0 to Cnt do
            fImgBuffer^[P + I] := Bt;
          Inc(J);
          Inc(P, Cnt + 1);
        end;
    until J >= Count;
    if Assigned(fSeedRow) then
      Move(fImgBuffer^[fRowCount * fImgBytes], fSeedRow^[0], fImgBytes);
  end;
  {-----------------------------------------}
  // Get Delta encoded row of an image data
  procedure GetDeltaRow(Count : integer);
  var
    Cnt, Cnt1, Bt : byte;
    I, J, Ofst : integer;
  begin
    if not Assigned(fSeedRow) then
      begin
        for I := 1 to Count do
          GetChar;
        Exit;
      end;
    if Count = 0 then
      begin
        Move(fSeedRow^[0], fImgBuffer^[fRowCount * fImgBytes], fImgBytes);
        Exit;
      end;
    J := -1;
    repeat
      Bt := byte(GetChar);
      Inc(J);
      Cnt := ((Bt ShR 5) and $07) + 1;
      Cnt1 := Bt and $1F;
      if Cnt1 = 31 then
        begin
          Ofst := 31;
          repeat
            Cnt1 := byte(GetChar);
            Ofst := Ofst + Cnt1;
            Inc(J);
          until (Cnt1 < 255) or (J = Count);
          if J = Count then
            Exit;
        end
      else
        Ofst := Cnt1;
      if Ofst > fImgBytes then
        begin
          for I := J + 1 to Count do
            GetChar;
          Exit;
        end;
      for I := 0 to Cnt - 1 do
        begin
          fSeedRow^[fLastByte + Ofst + I] := byte(GetChar);
          Inc(J);
        end;
      fLastByte := fLastByte + Ofst + Cnt;
    until J >= Count;
    Move(fSeedRow^[0], fImgBuffer^[fRowCount * fImgBytes], fImgBytes);
  end;
  {-----------------------------------------}
  // Run length encoding
  procedure GetRLERow(Count : integer);
  var
    Bt, Cnt : byte;
    I, N, K : integer;
  begin
    if Count = 0 then
      begin
        if Assigned(fSeedRow) then
          Move(fSeedRow^[0], fImgBuffer^[fRowCount * fImgBytes], fImgBytes);
        Exit;
      end;
    if Odd(Count) then
      begin
        for I := 1 to Count do
          GetChar;
        Exit;
      end;
    N := 1;
    for I := 1 to (Count div 2) do
      begin
        Cnt := byte(GetChar);
        Bt := byte(GetChar);
        if fDoImage then
          for K := 0 to Cnt do
            begin
              if N <= fImgBytes then
                fImgBuffer^[fRowCount * fImgBytes + N] := Bt;
              Inc(N);
            end;
      end;
    if Assigned(fSeedRow) then
      Move(fImgBuffer^[fRowCount * fImgBytes], fSeedRow^[0], fImgBytes);
  end;
  {-----------------------------------------}
  // Image row with no compression at all
  procedure GetImageRow(Count : integer);
  var
    I  : integer;
  begin
    if Count = 0 then
      begin
        if Assigned(fSeedRow) then
          Move(fSeedRow^[0], fImgBuffer^[fRowCount * fImgBytes], fImgBytes);
        Exit;
      end;
    for I := 0 to Count - 1 do
      if fDoImage and (I <= fImgBytes) then
        fImgBuffer^[fRowCount * fImgBytes + I] := byte(GetChar);
    if Assigned(fSeedRow) then
      Move(fImgBuffer^[fRowCount * fImgBytes], fSeedRow^[0], fImgBytes);
  end;
  {------------------------------------------}
  // Get image block, i. e. entire image in one shot
  procedure GetImageBlock(Count : integer);
  var
    I : integer;
  begin
//    fRTLHeight := GetInt;
    fRTLWidth := Byte(GetChar);
    for I := 1 to 3 do
      fRTLWidth := (fRTLWidth ShL 8) + Byte(GetChar);
    if fRTLWidth = 0 then
      begin
        for I := 1 to Count - 4 do
          GetChar;
        Exit;
      end;
    if fRtlHeight = 0 then
      begin
        I := fRtlWidth div 8;
        if (fRtlWidth mod 8) > 0 then
          Inc(I);
        fRTLHeight := (Count - 4) div I;
      end;
    FreeImageMemory;
    GetImageMemory;
    if fDoImage then
      begin
        Dec(Count, 4);
        for I := 0 to Count - 1 do
          fImgBuffer^[I] := byte(GetChar);
      end;
  end;
  {------------------------------------------}
begin
  Result := 0;
  Ch := PeekChar;
  if Ch = '.' then
    begin
      repeat
        Ch := GetChar;
      until (Ch = ':') or EOS;
      Exit;
    end;
  GetRTLString;
  if fRtlCmd = '' then
    Exit;
(*
{$IFDEF DEBUG}
{$I-}
  AssignFile(Tf, 'rtl.txt');
  if FileExists('rtl.txt') then
    Append(Tf)
  else
    Rewrite(Tf);
  if IOResult = 0 then
    begin
      WriteLN(Tf, fRTLCmd);
      CloseFile(Tf);
    end;
{$I+}
{$ENDIF}
*)
  Ch := fRtlCmd[1];
  case Ch of
    '%' : begin
            St1 := '';
            I := 1;
            repeat
              Inc(I);
              Ch := fRtlCmd[I];
              if IsNumber(Ch) then
                St1 := St1 + UpCase(Ch);
            until not IsNumber(Ch);
            Ch := UpCase(Ch);
            if St1 = '' then
              Exit;
            try
              N := StrToInt(St1);
            except
              N := $FF;
            end;
            case Ch of
              'A' : begin
                      fRTLMode := True;
                      if N = 1 then
                        begin
                          fRTLCap.X := Round(fPenPos.Y / fRTLScale);
                          fRTLCap.Y := Round(fPenPos.X / fRTLScale);
                        end;
                    end;
              'B' : begin
//                      IsHPGL2 := True;
                      if (N >= 0) and fPclFrame then
                        fIgnorePs := True;
                      fRTLMode := False;
                      if N  = 1 then
                        begin
                          fPenPos.X := Round(fRTLCap.Y * fRTLScale);
                          fPenPos.Y := Round(fRTLCap.X * fRTLScale);
                        end;
                    end;
              'X' : if N <> -12345 then
                      begin

                      end;
            end;
          end;
    '&' : begin
            Ch := fRtlCmd[2];
            if Ch = 'b' then
              begin
                I := 3;
                St1 := '';
                while fRtlCmd[I] <> 'W' do
                  begin
                    St1 := St1 + fRtlCmd[I];
                    Inc(I);
                  end;
                try
                  Y := StrToInt(St1);
                except
                  Y := 0;
                end;
                for N := 1 to Y do
                  GetChar;
              end
            else
              begin
                I := 3;
                repeat
                  St1 := '';
                  if fRtlCmd[1] = '+' then
                    K := 0
                  else if fRtlCmd[1] = '-' then
                    K := 1
                  else
                    K := 2;
                  if fRtlCmd[1] in ['+', '-'] then
                    Delete(fRtlCmd, 1, 1);
                  while IsNumber(fRtlCmd[I]) do
                    begin
                      St1 := St1 + fRtlCmd[I];
                      Inc(I);
                    end;
                  Ch := UpCase(fRtlCmd[I]);
                  case Ch of
                    'H' : begin
                            try
                              N := StrToInt(St1);
                            except
                              N := 0;
                            end;
                            if N > 0 then
                              N := Round(N * fRTLRes / 720);
                            case K of
                              0 : fRTLCap.X := fRTLCap.X + N;
                              1 : fRTLCap.X := fRTLCap.X - N;
                              2 : fRTLCap.X := N;
                            end;
                          end;
                    'N' : begin
                          end;
                  end;
                until (fRtlCmd[I] = 'N') or (fRtlCmd[I] = 'H') or EOS or fAborted;
              end;
          end;
    '*' : begin
            I := 2;
            repeat
              Ch := fRtlCmd[I];
              Application.ProcessMessages;
              case Ch of
                'b' : begin
                        repeat
                          St1 := '';
                          Inc(I);
                          while IsNumber(fRtlCmd[I]) do
                            begin
                              St1 := St1 + fRtlCmd[I];
                              Inc(I);
                            end;
                          Ch := UpCase(fRtlCmd[I]);
                          try
                            N := StrToInt(St1);
                          except
                            N := 0;
                          end;
                          case Ch of
                            'L' : ; //Direction := N = 0;
                            'M' : fEncoding := N;
                            'V' : for Y := 1 to N do
                                    GetChar;
                            'W' : begin
                                    if not fRasterMode then
                                      begin
                                        if fOldCount > 0 then
                                          Y := fOldCount
                                        else
                                          Y := fImgCount;
                                        FreeImageMemory;
                                        if fOldBytes > 0 then
                                          Y := fOldBytes
                                        else
                                          Y := fImgBytes;
                                        if fEncoding < 4 then
                                          GetImageMemory
                                        else
                                          fRasterMode := True;
                                        fRowCount := 0;
                                      end;
                                    if fRasterMode and
                                       (fImageParams.BitsNdx = 1) and
                                       ((fEncoding = 4) or
                                        ((fEncoding < 4) and
                                         (fRowCount < fRtlHeight))) then
                                      begin
                                        fLastByte := 1;
                                        case fEncoding of
                                          0 : GetImageRow(N);
                                          1 : GetRLERow(N);
                                          2 : GetTIFFRow(N);
                                          3 : GetDeltaRow(N);
                                          4 : GetImageBlock(N);
                                          else begin
                                            for Y := 1 to N do
                                              GetChar;
                                          end;
                                        end;
                                      end
                                    else
                                      for Y := 1 to N do
                                        begin
                                          GetChar;
                                          if EOS or fAborted then
                                            Break;
                                        end;
                                      Inc(fRowCount);
                                  end;
                            'Y' : begin
                                    if fRasterMode then
                                      begin
                                        if Assigned(fSeedRow) then
                                          FillChar(fSeedRow[0], fImgBytes, #0);
                                        if N = 1 then
                                          Inc(fRowCount);
                                      end
                                    else
                                      fRTLCap.Y := fRTLCap.Y + N;
                                  end;
                          end;
                        until ((fRtlCmd[I] >= 'L') and (fRtlCmd[I] <= 'Y')) or
                              EOS or fAborted;
                      end;
                'c' : begin
                        repeat
                          St1 := '';
                          Inc(I);
                          while IsNumber(fRtlCmd[I]) do
                            begin
                              St1 := St1 + fRtlCmd[I];
                              Inc(I);
                            end;
                          try
                            N := StrToInt(St1);
                          except
                            N := 0;
                          end;
                          Ch := UpCase(fRtlCmd[I]);
                          case Ch of
                            'G' : ;
                            'Q' : ;
                            'W' : begin
                                    for Y := 1 to N do
                                      GetChar;
                                  end;
                            'X' : begin
                                    try
                                      X := StrToInt(St1);
                                    except
                                      X := 0;
                                    end;
                                    fPage.X := Round(X * 1.411);
                                    fPclFrame := True;
                                    if fPage.X < fPage.Y then
                                      fRotation := 90
                                    else
                                      fRotation := 0;
                                  end;
                            'Y' : begin
                                    try
                                      Y := StrToInt(St1);
                                     except
                                      Y := 0;
                                    end;
                                    fPage.Y := Round(Y * 1.411);
                                    fPclFrame := True;
                                    if fPage.X < fPage.Y then
                                      fRotation := 90
                                    else
                                      fRotation := 0;
                                  end;
                          end;
                        until ((fRtlCmd[I] >= 'G') and (fRtlCmd[I] <= 'Y')) or
                              EOS or fAborted or (I > Length(fRtlCmd));
                      end;
                'l' : begin
                        St1 := '';
                        Inc(I);
                        while IsNumber(fRtlCmd[I]) do
                          begin
                            St1 := St1 + fRtlCmd[I];
                            Inc(I);
                          end;
//                        Ch := fRtlCmd[I];
                      end;
                'p' : repeat
                        St1 := '';
                        Inc(I);
                        while IsNumber(fRtlCmd[I]) do
                          begin
                            St1 := St1 + fRtlCmd[I];
                            Inc(I);
                          end;
                        if St1[1] = '+' then
                          K := 1
                        else if St1[1] = '-' then
                          K := -1
                        else
                          K := 0;
                        if fRtlCmd[1] in ['+', '-'] then
                          Delete(fRtlCmd, 1, 1);
                        if St1 = '' then
                          St1 := '0';
                        try
                          N := StrToInt(St1);
                        except
                          N := 0;
                        end;
                        Ch := UpCase(fRtlCmd[I]);
                        case Ch of
                          'P' : ;
                          'R' : ;
                          'X' : begin
                                  if fRasterMode then
                                    begin
                                      if fOldCount > 0 then
                                        X := fOldCount
                                      else
                                        X := fImgCount;
                                      if fOldBytes > 0 then
                                        Y := fOldBytes
                                      else
                                        Y := fImgBytes;
                                      FreeImageMemory;
                                      fRasterMode := False;
                                      fRowCount := 0;
                                      fImgBytes := 0;
                                    end;
                                  if K = 0 then
                                    fRTLCap.X := N
                                  else
                                    fRTLCap.X := fRTLCap.X + N * K;
                                end;
                          'Y' : begin
                                  if fRasterMode then
                                    begin
                                      if fOldCount > 0 then
                                        X := fOldCount
                                      else
                                        X := fImgCount;
                                      if fOldBytes > 0 then
                                        Y := fOldBytes
                                      else
                                        Y := fImgBytes;
                                      FreeImageMemory;
                                      fRasterMode := False;
                                      fRowCount := 0;
                                      fImgBytes := 0;
                                    end;
                                  if K = 0 then
                                    fRTLCap.Y := N
                                  else
                                    fRTLCap.Y := fRTLCap.Y + K * N;
                                end;
                        end;
                      until ((fRtlCmd[I] >= 'P') and (fRtlCmd[I] <= 'Y')) or
                            EOS or fAborted or (I > Length(fRtlCmd));
                'r' : begin
                        repeat
                          St1 := '';
                          Inc(I);
                          while IsNumber(fRtlCmd[I]) do
                            begin
                              St1 := St1 + fRtlCmd[I];
                              Inc(I);
                            end;
                          Ch := UpCase(fRtlCmd[I]);
                          if Ch <> 'C' then
                            try
                              N := StrToInt(St1);
                            except
                              N := 0;
                            end
                          else
                            N := 0;
                          case Ch of
                            'A' : if not fRasterMode then
                                    begin
                                      if (N = 0) or (N = 3) then
                                        fRTLCap.X := 0;
                                      fRowCount := 0;
                                    end;
                            'C' : if fRasterMode then
                                    begin
                                      RotateImage;
                                      FreeImageMemory;
                                      fRowCount := 0;
                                      fImgBytes := 0;
                                      fRTLCap.Y := fRTLCap.Y + fRtlHeight;
                                      fRTLWidth := 0;
                                      fRTLHeight := 0;
                                      fRasterMode := False;
                                    end;
                            'S' : if not fRasterMode then
                                    fRtlWidth := N;
                            'T' : if not fRasterMode then
                                    fRtlHeight := N;
                            'U' : ;
                          end;
                        until ((fRtlCmd[I] >= 'A') and (fRtlCmd[I] <= 'T')) or
                              EOS or fAborted or (I > Length(fRtlCmd));
                      end;
                't' : repeat
                        St1 := '';
                        Inc(I);
                        while IsNumber(fRtlCmd[I]) do
                          begin
                            St1 := St1 + fRtlCmd[I];
                            Inc(I);
                          end;
                        try
                          N := StrToInt(St1);
                        except
                          N := 0;
                        end;
                        Ch := UpCase(fRtlCmd[I]);
                        case Ch of
                          'H' : ;
                          'J' : ;
                          'R' : if not fRasterMode then
                                  begin
                                    fRTLRes := N;
                                    fRTLScale := HPGLSteps / fRTLRes;
                                  end;
                          'V' : ;
                        end;
                      until ((fRtlCmd[I] >= 'H') and (fRtlCmd[I] <= 'V')) or
                            EOS or fAborted or (I > Length(fRtlCmd));
                'v' : begin
                        repeat
                          St1 := '';
                          Inc(I);
                          while IsNumber(fRtlCmd[I]) do
                            begin
                              St1 := St1 + fRtlCmd[I];
                              Inc(I);
                            end;
                          Ch := UpCase(fRtlCmd[I]);
                          try
                            N := StrToInt(St1);
                          except
                            N := 0;
                          end;
                          case Ch of
                            'A' : R := N;
                            'B' : G := N;
                            'C' : B := N;
                            'I' : if not fRasterMode then
                                    fPens[N].Color := RGB(R, G, B);
                            'N' : ;
                            'O' : ;
                            'S' : if not fRasterMode then
                                    fPen := N;
                            'T' : ;
                            'W' : if not fRasterMode then
                                    begin
                                      for Y := 1 to N do
                                        begin
                                          if Y <= 18 then
                                            IBuf[Y] := Byte(GetChar)
                                          else
                                            GetChar;
                                        end;
                                      if N > 18 then
                                        N := 18;
                                      Move(IBuf[1], fImageParams, N);
                                      if fImageParams.BitsNdx = 0 then
                                        fImageParams.BitsNdx := 1;
                                    end
                                  else
                                    for Y := 1 to N do
                                      GetChar;
                          end;
                        until ((fRtlCmd[I] >= 'A') and (fRtlCmd[I] <= 'W')) or
                               EOS or fAborted or (I > Length(fRtlCmd));
                      end;
                else begin
                       Inc(I);
                       while IsNumber(fRtlCmd[I]) do
                         begin
                           Inc(I);
//                           Ch := fRtlCmd[I];
                         end;
                     end;
              end;
            until ((fRtlCmd[I] >= 'A') and (fRtlCmd[I] <= 'Z')) or
                    EOS or fAborted;
          end;
    'E' : Exit;
   end;
end;

{------------------------------------------------------}
procedure THPGL2.Paint;
var
  Cr : TCursor;
  Rc : TRect;
begin
  if fSizing then
    Exit;
  Canvas.Brush.Color := fPaper;
  if fBorder then
    Canvas.Pen.Style := psSolid
  else
    Canvas.Pen.Style := psClear;
  Rc := Rect(0, 0, Self.Width - 1, Self.Height - 1);
  Canvas.Pen.Width := 1;
  Canvas.Rectangle(Rc);
  if fBorder then
    Frame3d(Canvas, Rc, clBtnFace, clBtnShadow, 2);
  if csDesigning in ComponentState then
    Exit;
  if fPlotReady then
    begin
      if not fElements.ViewReady then
        ReportError(11, 'Element view is not ready')
      else if fElements.Count = 0 then
        ReportError(11, 'Element list is empty')
      else
        begin
          Cr := Cursor;
          fDrawing := True;
          if Assigned(fOnPaint) then
            fOnPaint(Self);
          Cursor := crHourGlass;
          fElements.BackColor := Self.fPaper;
          fElements.DrawList;
          while fElements.Drawing do
            Application.ProcessMessages;
        end;
      Cursor := Cr;
      fDrawing := False;
      if Assigned(fAfterPaint) then
        fAfterPaint(Self);
    end;
end;
{------------------------------------------------------}
procedure THPGL2.SetBorder(const Value: boolean);
begin
  fBorder := Value;
  Refresh;
end;
{------------------------------------------------------}
procedure THPGL2.SetFrame(const Value: TColor);
begin
  fFrame := Value;
  Refresh;
end;
{------------------------------------------------------}
procedure THPGL2.SetPaperColor(const Value: TColor);
begin
  fPaper := Value;
  fElements.BackColor := fPaper;
  Refresh;
end;
{------------------------------------------------------}
procedure THPGL2.SetAutoPaint(const Value: boolean);
begin
  fAutoPaint := Value;
  if fAutoPaint and fPlotReady then
    Refresh;
end;
{------------------------------------------------------}
function THPGL2.ElementType(Index: integer): TElemType;
begin
  if Index < fElements.Count then
    Result := fElements.GetType(Index)
  else
    Result := etNone;
end;
{------------------------------------------------------}
// Set Aborted to True while reding or drawing plot only;
procedure THPGL2.SetAborted(const Value: boolean);
begin
  fAborted := Value;
  if fDrawing then
   Elements.Aborted := True;
end;
{------------------------------------------------------}
procedure THPGL2.SetIncrement(const Value: integer);
begin
  fIncrement := Value;
end;
{------------------------------------------------------}
function THPGL2.GetExtents: TPoint;
var
  X, Y : integer;
begin
  if fPlotReady then
    with fElements.RealBounds do
      begin
        X := Round(Abs(Right - Left) / HPGLSteps);
        Y := Round(Abs(Top - Bottom) / HPGLSteps);
        Result := Point(X, Y);
      end
  else
    Result := Point(0 ,0);
end;
{------------------------------------------------------}
function THPGL2.Extents: TRect;
begin
  Result := fElements.ListBounds;
end;
{------------------------------------------------------}
procedure THPGL2.SetDoZoom(const Value: boolean);
begin
  fDoZoom := Value;
end;
{------------------------------------------------------}
// Check if we have to turn on zoom and pan
procedure THPGL2.MouseDown(Button: TMouseButton; Shift: TShiftState;
                           X, Y: Integer);
begin
  inherited;
  if fDrawing or (not fPlotReady) then
    Exit;
  if (not fReading) and fDoZoom and (not fMouseDn) then
    begin
      if Shift = [ssLeft] then
        begin
          fMouseDn := True;
          fStartX := X;
          fStartY := Y;
          fLastX := X;
          fLastY := Y;
          Canvas.MoveTo(fStartX, fStartY);
          Canvas.Pen.Mode := pmNotXOR;
          if fViewMode = vmZoom then
            Canvas.Pen.Color := clBlack
          else
            Canvas.Pen.Color := clRed;
          Canvas.Brush.Color := clBlack;
          Canvas.Pen.Width := 1;
        end
      else
        begin
          fElements.InitView(Rect(0, 0, Width, Height));
          Invalidate;
        end;
    end;
  if Assigned(OnMouseDown) then
    OnMouseDown(Self, Button, Shift, X, Y);
end;
{------------------------------------------------------}
// Draw zoom box or pan line
procedure THPGL2.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if Assigned(OnMouseMove) then
    OnMouseMove(Self, Shift, X, Y);
  if fDrawing or (not fPlotReady) then
    Exit;
  if fMouseDn then
    case fViewMode of
      vmZoom : begin
                 Canvas.Rectangle(fStartX, fStartY, fLastX, fLastY);
                 fLastX := X;
                 fLastY := Y;
                 Canvas.Rectangle(fStartX, fStartY, fLastX, fLastY);
               end;
      vmPan  : begin
                 Canvas.MoveTo(fStartX, fStartY);
                 Canvas.LineTo(fLastX, fLastY);
                 fLastX := X;
                 fLastY := Y;
                 Canvas.MoveTo(fStartX, fStartY);
                 Canvas.LineTo(fLastX, fLastY);
               end;
    end;
end;
{------------------------------------------------------}
// If in zoom or pan mode, redraw a plot
procedure THPGL2.MouseUp(Button: TMouseButton; Shift: TShiftState;
                         X, Y: Integer);
var
  R : TRect;
  A, B  : integer;
begin
  inherited;
  if fDrawing or (not fPlotReady) then
    Exit;
  if fMouseDn then
    begin
      fLastX := X;
      fLastY := Y;
      Canvas.Pen.Color := clBlack;
      Canvas.Pen.Mode := pmCopy;
      Canvas.Pen.Style := psSolid;
      Canvas.Brush.Style := bsSolid;
      if fViewMode = vmZoom then
        fElements.SetViewFromScreen(Rect(fStartX, fStartY, fLastX, fLastY))
      else
        begin
          R := fElements.View;
          A := Round((fStartX - fLastX) / fElements.XScale);
          B := Round((fStartY - fLastY) / fElements.YScale);
          R.Left := R.Left + A;
          R.Top := R.Top + B;
          R.Right := R.Right + A;
          R.Bottom := R.Bottom + B;
          fElements.View := R;
        end;
      fMouseDn := False;
      Invalidate; //fElements.DrawList;
    end;
  if Assigned(OnMouseUp) then
    OnMouseUp(Self, Button, Shift, X, Y);
end;
{------------------------------------------------------}
procedure THPGL2.SetViewMode(const Value: TViewMode);
begin
  if Value <= vmPan then
    fViewMode := Value
  else
    fViewMode := vmPan;
end;
{------------------------------------------------------}
procedure THPGL2.Print;
var
  Vp, Vw, Pr : TRect;
  Cr         : TCursor;
  Trans      : boolean;
  Clr        : TColor;
begin
  if fDrawing or (not fPlotReady) then
    Exit;
  Vp := fElements.ViewPort;
  Vw := fElements.View;
  Cr := Cursor;
  Cursor := crHourglass;
  Printer.Title := fPlotFile;
  Clr := fElements.BackColor;
  fElements.BackColor := clWhite;
  Trans := fElements.UseTransMode;
  fElements.UseTransMode := False;
  Printer.BeginDoc;
  Pr.Left := GetDeviceCaps(Printer.Handle, PHYSICALOFFSETX);
  Pr.Top := GetDeviceCaps(Printer.Handle, PHYSICALOFFSETY);
  Pr.Right := GetDeviceCaps(Printer.Handle, PHYSICALWIDTH) - Pr.Left * 2 - 40;
  Pr.Bottom := GetDeviceCaps(Printer.Handle, PHYSICALHEIGHT) - Pr.Top * 2 - 40;
  fElements.Canvas := Printer.Canvas;
  fElements.InitView(Pr);
  fElements.DrawList;
  Printer.EndDoc;
  fElements.Canvas := Self.Canvas;
  fElements.BackColor := Clr;
  fElements.SetViewFromScreen(Rect(0, 0, Self.Width - 1, Self.Height - 1));
  fElements.UseTransMode := Trans;
  Cursor := Cr;
  Invalidate;
end;
{------------------------------------------------------}
end.
